import time
import asyncio
import threading
import os
import logging
import sys
import traceback
import queue
import concurrent.futures
from typing import Optional, Callable, Dict, Set, List, Any, Union, Tuple
from contextlib import contextmanager
from functools import wraps
from dataclasses import dataclass, field
from enum import Enum, auto
from datetime import datetime
from collections import defaultdict

# 로깅 설정
def setup_logging(
    level: int = logging.INFO,
    log_file: str = 'fnsignal.log',
    enable_console: bool = True,
    enable_file: bool = True,
    format_str: str = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
) -> None:
    """
    로깅 설정을 구성합니다.
    
    Args:
        level (int): 로깅 레벨
        log_file (str): 로그 파일 경로
        enable_console (bool): 콘솔 출력 활성화 여부
        enable_file (bool): 파일 출력 활성화 여부
        format_str (str): 로그 포맷 문자열
    """
    handlers = []
    
    if enable_console:
        handlers.append(logging.StreamHandler(sys.stdout))
    if enable_file:
        handlers.append(logging.FileHandler(log_file, encoding='utf-8'))
        
    logging.basicConfig(
        level=level,
        format=format_str,
        handlers=handlers
    )

# 기본 로깅 설정
setup_logging()
logger = logging.getLogger(__name__)

class SignalPriority(Enum):
    LOW = auto()
    NORMAL = auto()
    HIGH = auto()
    CRITICAL = auto()

@dataclass
class SignalCallback:
    callback: Callable
    sender: Optional[str]
    priority: SignalPriority
    is_async: bool
    filter_condition: Optional[Callable[[Any], bool]]
    tags: Set[str] = field(default_factory=set)
    group: Optional[str] = None
    channel: Optional[str] = None
    stop_propagation: bool = False
    registered_at: datetime = field(default_factory=datetime.now)
    execution_count: int = 0
    total_execution_time: float = 0.0
    max_execution_time: float = 0.0
    error_count: int = 0
    
    def update_stats(self, execution_time: float, error: bool = False) -> None:
        self.execution_count += 1
        self.total_execution_time += execution_time
        self.max_execution_time = max(self.max_execution_time, execution_time)
        if error:
            self.error_count += 1
    
    @property
    def average_execution_time(self) -> float:
        return self.total_execution_time / self.execution_count if self.execution_count > 0 else 0.0
    
    @property
    def error_rate(self) -> float:
        return self.error_count / self.execution_count if self.execution_count > 0 else 0.0

@dataclass
class SignalStats:
    total_signals: int = 0
    active_callbacks: int = 0
    last_signal_time: Optional[datetime] = None
    signal_counts: Dict[str, int] = field(default_factory=lambda: defaultdict(int))
    error_counts: Dict[str, int] = field(default_factory=lambda: defaultdict(int))
    execution_times: Dict[str, List[float]] = field(default_factory=lambda: defaultdict(list))
    total_execution_time: float = 0.0
    max_execution_time: float = 0.0
    
    def update_execution_time(self, execution_time: float) -> None:
        self.total_execution_time += execution_time
        self.max_execution_time = max(self.max_execution_time, execution_time)
    
    @property
    def average_execution_time(self) -> float:
        return self.total_execution_time / self.total_signals if self.total_signals > 0 else 0.0

class SignalManager:
    _instance = None
    _lock = threading.Lock()
    _initialized = False
    _event_loop = None
    _event_loop_lock = threading.Lock()
    _thread_local = threading.local()

    def __new__(cls):
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        if not self._initialized:
            with self._lock:
                if not self._initialized:
                    self._callbacks: Dict[str, List[SignalCallback]] = defaultdict(list)
                    self._tag_index: Dict[str, Set[Tuple[str, int]]] = defaultdict(set)
                    self._group_index: Dict[str, Set[Tuple[str, int]]] = defaultdict(set)
                    self._channel_index: Dict[str, Set[Tuple[str, int]]] = defaultdict(set)
                    self._stats = SignalStats()
                    self._signal_queue = queue.PriorityQueue()
                    self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=10)
                    self._running = True
                    self._print_lock = threading.Lock()
                    self._callback_locks: Dict[str, threading.Lock] = defaultdict(threading.Lock)
                    
                    # 메인 이벤트 루프 생성 및 별도 스레드에서 실행
                    self._event_loop = asyncio.new_event_loop()
                    def start_loop(loop):
                        asyncio.set_event_loop(loop)
                        loop.run_forever()
                    self._event_loop_thread = threading.Thread(target=start_loop, args=(self._event_loop,), daemon=True)
                    self._event_loop_thread.start()
                    
                    # 시그널 처리 스레드 시작
                    self._processing_thread = threading.Thread(target=self._process_signals, daemon=True)
                    self._processing_thread.start()
                    self._initialized = True

    def _process_signals(self):
        """시그널 처리 메인 루프"""
        # 이벤트 루프 설정
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        
        while self._running:
            try:
                priority_value, signal_data = self._signal_queue.get(timeout=0.1)
                signal_name, data, sender, tags, group, channel = signal_data
                
                # 각 시그널에 대한 락 획득
                with self._callback_locks[signal_name]:
                    # 이벤트 루프에서 콜백 실행
                    future = asyncio.run_coroutine_threadsafe(
                        self._execute_callbacks(signal_name, data, sender, tags, group, channel),
                        self._event_loop
                    )
                    try:
                        # 콜백 실행 완료 대기
                        future.result(timeout=30)  # 30초 타임아웃
                        self._stats.total_signals += 1
                        self._stats.signal_counts[signal_name] += 1
                        self._stats.last_signal_time = datetime.now()
                    except concurrent.futures.TimeoutError:
                        logger.error(f"시그널 처리 타임아웃: {signal_name}")
                    except Exception as e:
                        logger.error(f"시그널 처리 오류: {e}")
                        logger.error(traceback.format_exc())
                
                self._signal_queue.task_done()
            except queue.Empty:
                continue
            except Exception as e:
                logger.error(f"시그널 처리 오류: {e}")
                logger.error(traceback.format_exc())
        
        # 이벤트 루프 정리
        loop.close()

    async def _execute_callbacks(self, signal_name: str, data: Any, sender: Any, tags: Set[str], group: str, channel: str):
        """콜백 실행을 위한 비동기 메서드"""
        logger.info(f"콜백 실행 시작: {signal_name} - 데이터: {data}")
        
        # 콜백 목록의 안전한 복사본 생성 및 정렬
        with self._lock:
            callbacks = [
                callback for callback in self._callbacks.get(signal_name, [])
                if (callback.sender is None or callback.sender == sender) and
                   (not tags or callback.tags.intersection(tags)) and
                   (not group or callback.group == group) and
                   (not channel or callback.channel == channel)
            ]
            logger.info(f"찾은 콜백 수: {len(callbacks)}")
            
            # 우선순위에 따라 정렬
            callbacks.sort(key=lambda x: {
                SignalPriority.LOW: 3,
                SignalPriority.NORMAL: 2,
                SignalPriority.HIGH: 1,
                SignalPriority.CRITICAL: 0
            }[x.priority])

        # 각 콜백 실행
        for callback in callbacks:
            try:
                logger.info(f"콜백 실행 시도: {callback.callback.__name__} (is_async={callback.is_async})")
                start_time = time.time()
                
                if callback.is_async:
                    # 비동기 콜백 실행
                    result = await callback.callback(signal_name, data, sender)
                else:
                    # 동기 콜백 실행
                    result = callback.callback(signal_name, data, sender)
                
                execution_time = time.time() - start_time
                callback.update_stats(execution_time)
                self._stats.update_execution_time(execution_time)
                logger.info(f"콜백 실행 완료: {callback.callback.__name__} - 실행 시간: {execution_time:.3f}초")
                
                if callback.stop_propagation:
                    logger.info("시그널 전파 중단")
                    break
                    
            except Exception as e:
                logger.error(f"콜백 실행 오류: {e}")
                logger.error(traceback.format_exc())
                callback.update_stats(0, error=True)
                self._stats.error_counts[signal_name] += 1

    def _update_index(self, index_dict: Dict[str, Set[Tuple[str, int]]], key: str, signal_name: str, callback_index: int, add: bool = True):
        """인덱스 업데이트를 위한 헬퍼 메서드"""
        if add:
            index_dict[key].add((signal_name, callback_index))
        else:
            index_dict[key].discard((signal_name, callback_index))
            # 빈 세트 제거
            if not index_dict[key]:
                del index_dict[key]

    def _remove_from_index(self, index_dict: Dict[str, Set[Tuple[str, int]]], signal_name: str, indices: Set[int]):
        """인덱스에서 항목 제거를 위한 헬퍼 메서드"""
        keys_to_remove = set()
        for key, entries in index_dict.items():
            # 제거할 항목 찾기
            to_remove = {(signal_name, idx) for idx in indices if (signal_name, idx) in entries}
            if to_remove:
                entries -= to_remove
                if not entries:  # 빈 세트인 경우
                    keys_to_remove.add(key)
        
        # 빈 세트 제거
        for key in keys_to_remove:
            del index_dict[key]

    def register_callback(
        self,
        signal_name: str,
        callback: Callable,
        sender: Any = None,
        priority: SignalPriority = SignalPriority.NORMAL,
        filter_condition: Optional[Callable] = None,
        tags: Set[str] = None,
        group: Optional[str] = None,
        channel: Optional[str] = None,
        stop_propagation: bool = False,
        is_async: Optional[bool] = None,
    ):
        """콜백 등록 (스레드 안전)"""
        with self._lock:
            # 이미 등록된 콜백인지 확인
            existing_callbacks = self._callbacks.get(signal_name, [])
            for existing_callback in existing_callbacks:
                if (existing_callback.callback == callback and 
                    existing_callback.sender == sender and 
                    existing_callback.filter_condition == filter_condition):
                    logger.warning(f"콜백이 이미 등록되어 있습니다: {signal_name} - {callback.__name__}")
                    return

            # is_async를 명시적으로 받도록
            if is_async is None:
                is_async = asyncio.iscoroutinefunction(callback)
            callback_obj = SignalCallback(
                callback=callback,
                sender=sender,
                priority=priority,
                is_async=is_async,
                filter_condition=filter_condition,
                tags=tags or set(),
                group=group,
                channel=channel,
                stop_propagation=stop_propagation
            )
            
            # 콜백 등록
            self._callbacks[signal_name].append(callback_obj)
            callback_index = len(self._callbacks[signal_name]) - 1
            
            # 인덱스 업데이트
            for tag in callback_obj.tags:
                self._update_index(self._tag_index, tag, signal_name, callback_index)
            if group:
                self._update_index(self._group_index, group, signal_name, callback_index)
            if channel:
                self._update_index(self._channel_index, channel, signal_name, callback_index)
            
            self._callbacks[signal_name].sort(key=lambda x: x.priority.value, reverse=True)
            self._stats.active_callbacks += 1
            logger.info(f"콜백 등록 완료: {signal_name} - {callback.__name__}")

    def unregister_callback(self, signal_name: str, callback: Callable, sender: Any = None):
        """콜백 제거 (스레드 안전)"""
        with self._lock:
            if signal_name in self._callbacks:
                original_count = len(self._callbacks[signal_name])
                removed_indices = set()
                
                # 제거할 콜백 찾기
                for i, cb in enumerate(self._callbacks[signal_name]):
                    if cb.callback == callback and (sender is None or cb.sender == sender):
                        removed_indices.add(i)
                
                if removed_indices:
                    # 인덱스에서 제거
                    self._remove_from_index(self._tag_index, signal_name, removed_indices)
                    self._remove_from_index(self._group_index, signal_name, removed_indices)
                    self._remove_from_index(self._channel_index, signal_name, removed_indices)
                    
                    # 콜백 제거
                    self._callbacks[signal_name] = [
                        cb for i, cb in enumerate(self._callbacks[signal_name])
                        if i not in removed_indices
                    ]
                    
                    # 빈 리스트 제거
                    if not self._callbacks[signal_name]:
                        del self._callbacks[signal_name]
                    
                    removed_count = original_count - len(self._callbacks.get(signal_name, []))
                    self._stats.active_callbacks -= removed_count
                    if removed_count > 0:
                        logger.info(f"콜백 제거 완료: {signal_name} - {callback.__name__} ({removed_count}개)")

    def get_callbacks_by_tag(self, tag: str) -> List[Tuple[str, SignalCallback]]:
        """태그로 콜백을 검색합니다."""
        with self._lock:
            return [
                (signal_name, self._callbacks[signal_name][idx])
                for signal_name, idx in self._tag_index.get(tag, set())
            ]

    def get_callbacks_by_group(self, group: str) -> List[Tuple[str, SignalCallback]]:
        """그룹으로 콜백을 검색합니다."""
        with self._lock:
            return [
                (signal_name, self._callbacks[signal_name][idx])
                for signal_name, idx in self._group_index.get(group, set())
            ]

    def get_callbacks_by_channel(self, channel: str) -> List[Tuple[str, SignalCallback]]:
        """채널로 콜백을 검색합니다."""
        with self._lock:
            return [
                (signal_name, self._callbacks[signal_name][idx])
                for signal_name, idx in self._channel_index.get(channel, set())
            ]

    def send_signal(
        self,
        signal_name: str,
        data: Any = None,
        sender: Any = None,
        priority: SignalPriority = SignalPriority.NORMAL,
        tags: Set[str] = None,
        group: Optional[str] = None,
        channel: Optional[str] = None
    ):
        """시그널을 전송합니다."""
        if not self._running:
            raise RuntimeError("시그널 매니저가 종료되었습니다.")
            
        # 우선순위 값을 정수로 변환
        priority_value = {
            SignalPriority.LOW: 3,
            SignalPriority.NORMAL: 2,
            SignalPriority.HIGH: 1,
            SignalPriority.CRITICAL: 0
        }[priority]
        
        # 시그널 데이터를 큐에 추가
        signal_data = (signal_name, data, sender, tags or set(), group, channel)
        self._signal_queue.put((priority_value, signal_data))
        
        logger.debug(f"시그널 전송: {signal_name} (우선순위: {priority.name})")

    def get_signal_stats(self) -> SignalStats:
        return self._stats

    def reset_signal_stats(self):
        with self._lock:
            self._stats = SignalStats()

    def shutdown(self):
        """시그널 매니저 종료"""
        self._running = False
        if self._processing_thread.is_alive():
            self._processing_thread.join(timeout=5)
        self._executor.shutdown(wait=True)
        if self._event_loop and self._event_loop.is_running():
            self._event_loop.call_soon_threadsafe(self._event_loop.stop)
            self._event_loop_thread.join(timeout=5)
            self._event_loop.close()

    def on(
        self,
        signal_name: str,
        sender: Any = None,
        priority: SignalPriority = SignalPriority.NORMAL,
        filter_condition: Optional[Callable] = None,
        tags: Set[str] = None,
        group: Optional[str] = None,
        channel: Optional[str] = None,
        stop_propagation: bool = False
    ):
        """
        시그널 핸들러를 등록하기 위한 데코레이터입니다.
        
        Args:
            signal_name (str): 시그널 이름
            sender (Any, optional): 시그널 발신자
            priority (SignalPriority, optional): 시그널 우선순위
            filter_condition (Callable, optional): 시그널 필터링 조건
            tags (Set[str], optional): 시그널 태그
            group (str, optional): 시그널 그룹
            channel (str, optional): 시그널 채널
            stop_propagation (bool, optional): 시그널 전파 중단 여부
        """
        def decorator(func: Callable):
            is_async = asyncio.iscoroutinefunction(func)
            if is_async:
                @wraps(func)
                async def async_wrapper(*args, **kwargs):
                    return await func(*args, **kwargs)
                wrapped_func = async_wrapper
            else:
                wrapped_func = func

            self.register_callback(
                signal_name=signal_name,
                callback=wrapped_func,
                sender=sender,
                priority=priority,
                filter_condition=filter_condition,
                tags=tags,
                group=group,
                channel=channel,
                stop_propagation=stop_propagation,
                is_async=is_async
            )
            return wrapped_func
        return decorator

    def on_tag(
        self,
        tag: str,
        signal_name: str = None,
        sender: Any = None,
        priority: SignalPriority = SignalPriority.NORMAL,
        filter_condition: Optional[Callable] = None,
        group: Optional[str] = None,
        channel: Optional[str] = None,
        stop_propagation: bool = False
    ):
        """
        특정 태그를 가진 시그널 핸들러를 등록하기 위한 데코레이터입니다.
        
        Args:
            tag (str): 시그널 태그
            signal_name (str, optional): 시그널 이름
            sender (Any, optional): 시그널 발신자
            priority (SignalPriority, optional): 시그널 우선순위
            filter_condition (Callable, optional): 시그널 필터링 조건
            group (str, optional): 시그널 그룹
            channel (str, optional): 시그널 채널
            stop_propagation (bool, optional): 시그널 전파 중단 여부
        """
        def decorator(func: Callable):
            is_async = asyncio.iscoroutinefunction(func)
            if is_async:
                @wraps(func)
                async def async_wrapper(*args, **kwargs):
                    return await func(*args, **kwargs)
                wrapped_func = async_wrapper
            else:
                wrapped_func = func

            self.register_callback(
                signal_name=signal_name or func.__name__,
                callback=wrapped_func,
                sender=sender,
                priority=priority,
                filter_condition=filter_condition,
                tags={tag},
                group=group,
                channel=channel,
                stop_propagation=stop_propagation,
                is_async=is_async
            )
            return wrapped_func
        return decorator

    def on_group(
        self,
        group: str,
        signal_name: str = None,
        sender: Any = None,
        priority: SignalPriority = SignalPriority.NORMAL,
        filter_condition: Optional[Callable] = None,
        tags: Set[str] = None,
        channel: Optional[str] = None,
        stop_propagation: bool = False
    ):
        """
        특정 그룹의 시그널 핸들러를 등록하기 위한 데코레이터입니다.
        
        Args:
            group (str): 시그널 그룹
            signal_name (str, optional): 시그널 이름
            sender (Any, optional): 시그널 발신자
            priority (SignalPriority, optional): 시그널 우선순위
            filter_condition (Callable, optional): 시그널 필터링 조건
            tags (Set[str], optional): 시그널 태그
            channel (str, optional): 시그널 채널
            stop_propagation (bool, optional): 시그널 전파 중단 여부
        """
        def decorator(func: Callable):
            is_async = asyncio.iscoroutinefunction(func)
            if is_async:
                @wraps(func)
                async def async_wrapper(*args, **kwargs):
                    return await func(*args, **kwargs)
                wrapped_func = async_wrapper
            else:
                wrapped_func = func

            self.register_callback(
                signal_name=signal_name or func.__name__,
                callback=wrapped_func,
                sender=sender,
                priority=priority,
                filter_condition=filter_condition,
                tags=tags,
                group=group,
                channel=channel,
                stop_propagation=stop_propagation,
                is_async=is_async
            )
            return wrapped_func
        return decorator
    def receive_signal(self, signal_name: str, timeout: Optional[float] = None) -> Any:
        """
        특정 시그널을 기다리고 수신합니다.
        
        Args:
            signal_name (str): 수신할 시그널 이름
            timeout (float, optional): 타임아웃 시간 (초)
            
        Returns:
            Any: 수신된 시그널 데이터
        """
        signal_queue = queue.Queue()
        
        def signal_handler(signal_name, data, sender):
            signal_queue.put(data)
            
        # 임시 콜백 등록
        self.register_callback(signal_name, signal_handler)
        
        try:
            # 시그널 대기
            return signal_queue.get(timeout=timeout)
        except queue.Empty:
            raise TimeoutError(f"시그널 수신 타임아웃: {signal_name}")
        finally:
            # 임시 콜백 제거
            self.unregister_callback(signal_name, signal_handler)

# 싱글톤 인스턴스 생성
signal_manager = SignalManager()